
// IDE: Arduino Leonardo

// kolejność na płytce, przewody płasko:
// (pierwszy pin od Arduino):
// DB9-3 - zielony - LP
// DB9-8 - niebieski - GND
// DB9-7 - fiolet - VCC
// DB9-6 - szary - FIRE
// (złącza cinch)

// uwaga! 20210510
// połączyłem DB9-3 (zielony) ->  4 (LP event), fototranzystor
// połączyłem DB9-6 (szary)   -> 16 (FIRE), guzik
// winno być odwrotnie
// ponieważ DB9-3 (LEFT) to guzik (zielony) -> 16
// ponieważ DB9-6 (FIRE) to foto  (szary)   -> 4
// zamienić miejscami tymczasowo szary i zielony na headerze: (licząc od Arduino): szary/niebieski/fiolet/zielony (cinch)

// we don't need burst after all

#include <AbsMouse.h>

// virtual screen
#define XMAX 1024
#define YMAX 640

// LM1881N
#define PIN_ODDEVEN 8         // PB4 - PCINT4
#define PIN_BURST   10        // PB6 - PCINT6
#define PIN_COMPOSITE_SYNC 9  // PB5 - PCINT5
#define PORT_COMPOSITE_SYNC PB5  // PB5 - PCINT5
#define PINPIN_COMPOSITE_SYNC PINB  // PB5 - PCINT5
#define PIN_VSYNC 7           // PE6 - INT6
// joyport
#define PIN_FIRE 16           // PB3 - PCINT3
#define PIN_LP 4              // PD4 - ICP1

volatile bool lpevent = false;
volatile bool odd = false;
volatile bool vsync = false;
volatile bool burst = false;
volatile bool lastFire = false;
volatile bool fire = false;
volatile uint32_t frame = 0; // runtime frame counter
volatile uint16_t lptime = 0; // lp event
volatile uint16_t lprow = 0; // screen line when LP happened
volatile uint16_t lprowtime = 0; // time when row started when LP happened
volatile uint16_t row = 0; // screen line counter
volatile uint16_t rowmax = 0; // number of lines on screen
volatile uint16_t lpwholerowtime = 0; // time when row ended when LP happened
volatile uint8_t previousPins = 0;

// boundaries for autocalibration, initially a rectangle in the middle of the screen
uint16_t xmin = XMAX/2-XMAX/4;
uint16_t xmax = XMAX/2+XMAX/4;
uint16_t ymin = YMAX/2-YMAX/4;
uint16_t ymax = YMAX/2+YMAX/4;

// smoothing
const uint8_t numReadings = 16;
uint8_t readIndex = 0;
uint32_t totalX = 0;
uint32_t totalY = 0;
uint16_t readingsX[numReadings];
uint16_t readingsY[numReadings];

ISR(PCINT0_vect) {
  uint8_t changedbits;
  changedbits = PINPIN_COMPOSITE_SYNC ^ previousPins;
  previousPins = PINPIN_COMPOSITE_SYNC;
  if (changedbits & (1<<PORT_COMPOSITE_SYNC)) {
    // only low to high triggers row count
    if ((previousPins & (1<<PORT_COMPOSITE_SYNC))) {
      uint16_t t = TCNT1;
      TCNT1 = 0; // reset timer counter for new row
      row++;
      if (lpevent && lpwholerowtime==0) {
        lpwholerowtime = t;
      }
    }
  }
}

ISR(TIMER1_CAPT_vect) {
  union twobyte {
    uint16_t word;
    uint8_t  byte[2];
  } timevalue;

  timevalue.byte[0] = ICR1L;    // grab captured timer value
  timevalue.byte[1] = ICR1H;    // grab captured timer value
  lptime = timevalue.word;
  lpevent = true;
  lprow = row;
  lpwholerowtime = 0;
}

void irq_vsync(void) {
  vsync = true;
  rowmax = row;
  row = 0;
  frame++;
  odd = digitalRead(PIN_ODDEVEN);
  burst = digitalRead(PIN_BURST);
  fire = !digitalRead(PIN_FIRE);
}

void initTimer(void) { 
  TCNT1 = 0;    //set the counter to 0, just in case  
  // Input Capture setup
  // ICNC1: Enable Input Capture Noise Canceler
  // ICES1: =1 for trigger on rising edge
  // CS11: =1 set prescaler to 8x system clock (F_CPU)
  // CS10: =1 set prescaler to 1x system clock (F_CPU)
  TCCR1A = 0;
  TCCR1B = (1<<ICNC1) | (0<<ICES1) | (1<<CS10); // enable noise canceler + 1x prescaler
  TCCR1C = 0;
  // Interrupt setup
  // ICIE1: Input capture
  // TOIE1: Timer1 overflow
  TIFR1 = (1<<ICF1) | (1<<TOV1);    // clear pending
  TIMSK1 = (1<<ICIE1) | (0<<TOIE1); // and enable
}

void setup() {
  //
  noInterrupts();
  // setup pins
  pinMode(PIN_ODDEVEN, INPUT);
  pinMode(PIN_BURST, INPUT);
  pinMode(PIN_COMPOSITE_SYNC, INPUT);
  pinMode(PIN_VSYNC, INPUT);
  pinMode(PIN_FIRE, INPUT_PULLUP);
  pinMode(PIN_LP, INPUT_PULLUP);
  // attach handlers
  attachInterrupt(digitalPinToInterrupt(PIN_VSYNC), irq_vsync, FALLING);
  //
  PCICR |= (1<<PCIE0);      // enable PCINT event
  PCMSK0 |= (1<<PCINT3) | (1<<PCINT5);  // interrupts only from those two pin-sources
  // setup Timer1
  initTimer();
  // setup volatile variables
  odd = false;
  vsync = false;
  lpevent = false;
  fire = false;
  lastFire = fire;
  frame = 0;
  lptime = 0;
  // setup smoothing
  for (uint8_t i = 0; i < numReadings; i++) {
    readingsX[i] = 0;
    readingsY[i] = 0;
  }
  //
  interrupts();
  //
  AbsMouse.init(XMAX,YMAX);
}

void loop() {
  if (vsync && lpevent) {
    if (lpevent) {
      uint16_t x = lptime; // 0..1024
      uint16_t y = lprow*2 + odd; // 0..640

      totalX = totalX - readingsX[readIndex];
      readingsX[readIndex] = x;
      totalX = totalX + readingsX[readIndex];
      totalY = totalY - readingsY[readIndex];
      readingsY[readIndex] = y;
      totalY = totalY + readingsY[readIndex];
      readIndex++;
      if (readIndex>=numReadings) {
        readIndex = 0;
      }

      x = totalX / numReadings;
      y = totalY / numReadings;

      x = constrain(x, 0, XMAX);  // constrain after smoothing
      y = constrain(y, 0, YMAX);

      if (fire) { // calibrate only when button pressed, don't pass the press event further
        if (x<xmin && x>0) { fire=false; xmin=x; } else { if (x>xmax && x<XMAX) { fire=false; xmax=x; } };
        if (y<ymin && y>0) { fire=false; ymin=y; } else { if (y>ymax && y<YMAX) { fire=false; ymax=y; } };
      }

      x = map(x, xmin, xmax, 0, XMAX); // rescale from [xmin..xmax] to [0..XMAX]
      y = map(y, ymin, ymax, 0, YMAX);
      x = constrain(x, 0, XMAX);  // constrain after scaling
      y = constrain(y, 0, YMAX);

      AbsMouse.move(x,y);
    }
    if (fire!=lastFire) {
      if (fire) {
        AbsMouse.press(MOUSE_LEFT);
      } else {
        AbsMouse.release(MOUSE_LEFT);
      }
      lastFire = fire;
    }
    // reset flags
    vsync = false;
    lpevent = false;
  }
}
